@@ -1,4 +1,4 @@
-module TestInit (module Export) where
+module TestInit (module Export, isWindows) where
 
 import Test.HUnit as Export hiding (path)
 import Test.Hspec as Export
@@ -8,3 +8,6 @@ import Shelly.Lifted as Export
 import Shelly as Export
 #endif
 import Test.Hspec.Contrib.HUnit ()
+import System.Info(os)
+isWindows :: Bool
+isWindows = os == "mingw32"
